#!/usr/bin/env python
# -*- coding: ASCII -*-

"""

:Author: Martin Kircher
:Contact: mkircher@uw.edu
:Date: *18.11.2014

"""

import sys, os
import pysam
import gzip
from optparse import OptionParser

sys.path.append(os.path.dirname(os.path.realpath(__file__)))
from AnalysisLib import get_from_tabix
from AnalysisLib import eval_dbSNP_frequencies

parser = OptionParser()
parser.add_option("-p","--path", dest="path", help="Path to scored variants (default './whole_genome_SNVs_anno.tsv.gz')",default="./whole_genome_SNVs_anno.tsv.gz")
parser.add_option("--found_out", dest="found_out", help="Write found variants to file (default 'output.tsv')",default="output.tsv")
parser.add_option("--header", dest="header", help="Write full header to output (default none)",default=False, action="store_true")
parser.add_option("--ignore_strand", dest="ignore_strand", help="Check both strands (default Off)",default=False,action="store_true")
parser.add_option("--no-indels", dest="noindels", help="Write found variants to stdout and do not have generate output file (default Off)",default=False, action="store_true")
(options, args) = parser.parse_args()

complement_base = { 'A':'T', 'C':'G', 'T':'A', 'G':'C' }

regionTabix = None

header = ["Chrom","Pos","Ref","Anc","Alt","Type","Length","isTv","isDerived","AnnoType","Consequence","ConsScore","ConsDetail","GC","CpG","mapAbility20bp","mapAbility35bp","scoreSegDup","priPhCons","mamPhCons","verPhCons","priPhyloP","mamPhyloP","verPhyloP","GerpN","GerpS","GerpRS","GerpRSpval","bStatistic","mutIndex","dnaHelT","dnaMGW","dnaProT","dnaRoll","mirSVR-Score","mirSVR-E","mirSVR-Aln","targetScan","fitCons","cHmmTssA","cHmmTssAFlnk","cHmmTxFlnk","cHmmTx","cHmmTxWk","cHmmEnhG","cHmmEnh","cHmmZnfRpts","cHmmHet","cHmmTssBiv","cHmmBivFlnk","cHmmEnhBiv","cHmmReprPC","cHmmReprPCWk","cHmmQuies","EncExp","EncH3K27Ac","EncH3K4Me1","EncH3K4Me3","EncNucleo","EncOCC","EncOCCombPVal","EncOCDNasePVal","EncOCFairePVal","EncOCpolIIPVal","EncOCctcfPVal","EncOCmycPVal","EncOCDNaseSig","EncOCFaireSig","EncOCpolIISig","EncOCctcfSig","EncOCmycSig","Segway","tOverlapMotifs","motifDist","motifECount","motifEName","motifEHIPos","motifEScoreChng","TFBS","TFBSPeaks","TFBSPeaksMax","isKnownVariant","ESP_AF","ESP_AFR","ESP_EUR","TG_AF","TG_ASN","TG_AMR","TG_AFR","TG_EUR","minDistTSS","minDistTSE","GeneID","FeatureID","CCDS","GeneName","cDNApos","relcDNApos","CDSpos","relCDSpos","protPos","relProtPos","Domain","Dst2Splice","Dst2SplType","Exon","Intron","oAA","nAA","Grantham","PolyPhenCat","PolyPhenVal","SIFTcat","SIFTval","RawScore","PHRED"]
fref,falt = 2,4

noTabix = False
if os.path.exists(options.path) and os.path.exists(options.path+".tbi"):
  filename = options.path
  sys.stderr.write("Opening %s...\n"%(filename))
  regionTabix = pysam.Tabixfile(filename,'r')
  infile = gzip.open(filename)
  for line in infile:
    if line.startswith("#"): continue
    fields = line.split()
    if len(fields) != len(header):
      header = ["Chrom","Pos","Ref","Alt","RawScore","PHRED"]
      fref,falt = 2,3
    break
else:
  sys.stderr.write("No valid file with pre-scored variants.\n")
  regionTabix = None
  noTabix = True

if options.noindels:
  found_out = sys.stdout
else:
  found_out = open(options.found_out,'w')

if options.header:
    found_out.write("#"+"\t".join(header)+"\n")

wroteStdout = False
for line in sys.stdin:
  if line.startswith('#'): continue
  if noTabix:
    if not options.noindels:
      sys.stdout.write(line)
      wroteStdout = True
  else:
    fields = line.rstrip().split('\t')
    chrom = fields[0]
    pos = int(fields[1])
    lref,allele = fields[-2],fields[-1]
    if lref == allele: continue

    found = False  
    for VEPline in regionTabix.fetch(chrom,pos-1,pos):
      vfields = VEPline.rstrip().split('\t')
      if (vfields[fref] == lref) and (vfields[falt] == allele):
        found_out.write(VEPline+"\n")
        found = True 
        break

    if not found and options.ignore_strand:
      for VEPline in regionTabix.fetch(chrom,pos-1,pos):
        vfields = VEPline.rstrip().split('\t')
        callele,clref = allele,lref
        if allele in complement_base: callele = complement_base[allele]
        if lref in complement_base: clref = complement_base[lref]
        for VEPline in hlines:
          if (VEPline[fref] == clref) and (VEPline[falt] == callele): 
            found_out.write("\t".join(VEPline)+"\n")
            found = True
            break
          elif (VEPline[fref] == allele) and (VEPline[falt] == lref): 
            found_out.write("\t".join(VEPline)+"\n")
            found = True
            break
          elif (VEPline[fref] == callele) and (VEPline[falt] == clref): 
            found_out.write(VEPline+"\n")
            found = True
            break                

    if not found:
      if not options.noindels:
        sys.stdout.write(line)
        wroteStdout = True

if not options.noindels:
  found_out.close()
  if not wroteStdout:
    sys.stdout.write("##EMPTY FILE\n")
